<?php
/**
 * Database library
 *
 * This file contains all the classes which take care of executing
 * operations on the database. They are all extensions of respective
 * PDO classes and also implement an error handling system. If a
 * critical error occurs, the execution will be stopped and, if the
 * debug mode is on, the error data will be shown.
 *
 * @package Database
 * @author  Alessandro Desantis <desa.alessandro@gmail.com>
 *
 * @link http://www.php.net/manual/en/book.pdo.php
 */

/**
 * Database Adapter
 *
 * This is the class which is used to connect and to perform every
 * kind of operation on the database.
 *
 * @package Database
 * @author  Alessandro Desantis <desa.alessandro@gmail.com>
 *
 * @link http://www.php.net/manual/en/class.pdo.php
 */
class Database extends PDO
{
    /**
     * Contains the database access data.
     *
     * @var    array
     * @access private
     */
    static private $_db = array();

    /**
     * Contains the instance of the class.
     *
     * @var    Database
     * @access private
     */
    static private $_instance = null;

    /**
     * Sets the database access data.
     *
     * @param array Data
     *
     * @access public
     */
    static public function setConnectionData($data)
    {
        self::$_db = $data;
    }

    /**
     * Returns the instance of the class.
     *
     * @return Database
     * @access public
     */
    static public function getInstance()
    {
        if (self::$_instance == null) {
            self::$_instance = new self();
        }

        return self::$_instance;
    }

    /**
     * Connects to the database and sets the statement class.
     *
     * @return Database
     * @access public
     */
    public function __construct()
    {
        global $config;

        try {
            parent::__construct(self::$_db['dsn'], self::$_db['user'], self::$_db['pass']);
            $this->setAttribute(PDO::ATTR_STATEMENT_CLASS, array('Database_Statement'));
        }
        catch (PDOException $e) {
            echo '<h1>Database error</h1>';
            echo '<p>An error occurred while connecting to the database.</p>';

            if (DEBUG_MODE) {
                $error = $e->getMessage();

                echo '<p>The server returned the following error:</p>';
                echo "<pre>{$error}</pre>";
            }

            exit();
        }
    }

    /**
     * Executes an SQL query on the database. This method actually
     * just creates a new statement with no placeholders and then
     * executes it, so that we can take advantage of the error
     * handling system implemented by the {@link Database_Statement}
     * class.
     *
     * @param string SQL code.
     *
     * @return Database_Statement
     * @access public
     */
    public function query($sql)
    {
        $stm = $this->prepare($sql);
        $stm->execute($sql);

        return $stm;
    }

    /**
     * Returns an object of type {@link Database_Record} or, if an
     * abstraction class exists, returns an instance of that class.
     *
     * @param string Table's name
     * @param string Table's key column
     * @param string Record's key
     *
     * @return object
     * @access public
     */
    public function getRecord($table, $column, $value)
    {
        $reflection_class = ucfirst(substr($table, 0, -1));

        if (class_exists($reflection_class)) {
            $reflection = new ReflectionClass($reflection_class);

            if ($reflection->isSubclassOf('Database_Record')) {
                return new $reflection_class($column, $value);
            }
        }

        return new Database_Record($table, $column, $value);
    }
}

/**
 * Database Statement
 *
 * This class represents an SQL statement which has to be executed on
 * the database. A statement can have one or more placeholder which,
 * during the execution, will be replaced with the real values.
 *
 * @package Database
 * @author  Alessandro Desantis <desa.alessandro@gmail.com>
 *
 * @link http://www.php.net/manual/en/class.pdostatement.php
 */
class Database_Statement extends PDOStatement
{
    /**
     * Executes the SQL statement eventually replacing the
     * placeholders with the specified values.
     *
     * @param array Values
     *
     * @access public
     */
    public function execute($values = array())
    {
        if (!parent::execute($values)) {
            echo '<h1>Database error</h1>';
            echo '<p>An error occurred while executing an SQL query on the database.</p>';

            if (DEBUG_MODE) {
                echo '<p>The SQL code of the query was the following:</p>';
                echo "<pre>{$this->queryString}</pre>";

                $error_data = $this->errorInfo();

                if (!empty($error_data[2])) {
                    echo '<p>The server returned the following error:</p>';
                    echo "<pre>{$error_data[2]}</pre>";
                }
            }

            exit();
        }
    }

    /**
     * Returns the data of a single row retrieved by a query.
     *
     * @return array
     * @access public
     */
    public function fetch()
    {
        return parent::fetch(PDO::FETCH_ASSOC);
    }

    /**
     * Returns an array containing the data of the entire rowset
     * retrieved by an SQL statement.
     *
     * @return array
     * @access public
     */
    public function fetchAll()
    {
        return parent::fetchAll(PDO::FETCH_ASSOC);
    }
}

/**
 * Database Record
 *
 * This class represents a single record in a table of the database
 * It's an abstraction layer: it is meant to provide a more real
 * approach to the application developing.
 *
 * @package Database
 * @author  Alessandro Desantis <desa.alessandro@gmail.com>
 *
 * @link http://www.php.net/manual/en/class.arrayaccess.php
 */
class Database_Record implements ArrayAccess
{
    /**
     * Contains the name of the table which contains the record.
     *
     * @var    string
     * @access protected
     */
    protected $table = '';

    /**
     * Contains the name of the column which contains the record's key.
     *
     * @var    string
     * @access protected
     */
    protected $column = '';

    /**
     * Contains the record's key (usually an integer).
     *
     * @var    mixed
     * @access protected
     */
    protected $value = '';

    /**
     * Contains the record's data.
     *
     * @var    array
     * @access protected
     */
    protected $data = array();

    /**
     * Gets the record's data from the table and saves them into
     * the instance.
     *
     * @param string    Table's name
     * @param string    Table's key column
     * @param mixed     Record's key
     *
     * @return Database_Record
     * @access public
     *
     * @throws DatabaseException If no record is found
     */
    public function __construct($table, $column, $value)
    {
        $this->table  = $table;
        $this->column = $column;
        $this->value  = $value;

        $stm = Database::getInstance()->prepare("SELECT * FROM `{$table}` WHERE `{$column}` = ?");
        $stm->execute(array($value));

        if (!$stm->rowCount()) {
            throw new DatabaseException($table, $column, $value);
        }

        $this->data = $stm->fetch();
    }

    /**
     * This magic method allows you to control the behaviour of
     * the class when the application tries to read a property
     * which does not exist. In this case we'll return the
     * respective column's value.
     *
     * @param string Property's name
     *
     * @return mixed  The value or null if the property does not exist.
     * @access public
     */
    public function __get($property)
    {
        if (!$this->__isset($property)) {
            return null;
        }

        $custom_method = 'get'. ucfirst($property);

        if (method_exists($this, $custom_method)) {
            return $this->$custom_method($this->data[$property]);
        }

        return $this->data[$property];
    }

    /**
     * This magic method allows you to control the behaviour of
     * the class when the application tries to set a property
     * which does not exist. In this case we'll save it so that
     * we can later write the changes to the table.
     *
     * @param string Property's name
     * @param mixed  Property's value
     *
     * @access public
     */
    public function __set($property, $value)
    {
        if (!$this->__isset($property)) {
            return;
        }

        $custom_method = 'set' . ucfirst($property);

        if (method_exists($this, $custom_method)) {
            $this->data[$property] = $this->$custom_method($value);
        }
        else {
            $this->data[$property] = $value;
        }
    }

    /**
     * This magic method allows you to control the behaviour of
     * the class when the application tries to check if a property
     * which does not exist is set. In this case we'll check it
     * the respective column exists in the table.
     *
     * @param string Property's name
     *
     * @return bool
     * @access public
     */
    public function __isset($property)
    {
        $custom_method = 'isset' . ucfirst($property);

        if (method_exists($this, $custom_method)) {
            return $this->$custom_method();
        }

        return isset($this->data[$property]);
    }

    /**
     * This magic method allows you to control the behaviour of
     * the class when the application tries to unset a property
     * which does not exist. In this case we'll unset the column
     * so that the database can automatically fill it.
     *
     * @param string Property's name
     *
     * @access public
     */
    public function __unset($property)
    {
        if (!$this->__isset($property)) {
            return;
        }

        $custom_method = 'unset' . ucfirst($property);

        if (method_exists($this, $custom_method)) {
            $this->$custom_method();
        }
        else {
            unset($this->data[$property]);
        }
    }

    /**
     * This magic method provides array access and is called when
     * the application tries to check if an offset exists. In this
     * case we'll check and return if the respective column exists.
     *
     * @param string Offset's name
     *
     * @return bool
     * @access public
     */
    public function offsetExists($offset)
    {
        return $this->__isset($this->$offset);
    }

    /**
     * This magic method provides array access and is called when
     * the application tries to get the value of an offset. In this
     * case we'll return the respective column's value.
     *
     * @param string Offset's name
     *
     * @return mixed  The value or null if the offset doesn't exist
     * @access public
     */
    public function offsetGet($offset)
    {
        return $this->__get($this->$offset);
    }

    /**
     * This magic method provides array access and is called when
     * the application tries to set the value of an offset. In this
     * case we'll set the respective column's value.
     *
     * @param string Offset's name
     * @param mixed  Offset's value
     *
     * @access public
     */
    public function offsetSet($offset, $value)
    {
        $this->__set($offset, $value);
    }

    /**
     * This magic method provides array access and is called whenm
     * the application tries to unset an offset. In this case we'll
     * unset the respective column's value.
     *
     * @param string Offset's name
     *
     * @access public
     */
    public function offsetUnset($offset)
    {
        $this->__unset($offset);
    }

    /**
     * Returns the record's data (applies all the modifiers).
     *
     * @return array
     * @access public
     */
    public function getData()
    {
        $data = array();

        foreach ($this->data as $column => $value) {
            $data[$column] = $this->__get($column);
        }

        return $data;
    }

    /**
     * Saves all the changes which were made to the record
     * to the database.
     *
     * @access public
     */
    public function commit()
    {
        $data = array();
        $sql  = "UPDATE `{$this->table}` SET ";

        foreach ($this->data as $column => $value) {
            $sql .= "`{$column}` = ?, ";
            $data[] = $value;
        }

        $sql = substr($sql, 0, -2);
        $sql .= " WHERE `{$this->column}` = ?";

        $data[] = $this->value;

        $stm = Database::getInstance()->prepare($sql);
        $stm->execute($data);
    }

    /**
     * Deletes the record from the database and clears the
     * saved data. THIS OPERATION CANNOT BE UNDONE.
     *
     * @access public
     */
    public function delete()
    {
        $stm = Database::getInstance()->prepare("DELETE FROM `{$this->table}` WHERE `{$this->column}` = ?");
        $stm->execute(array($this->value));

        $this->data = array();
    }
}

/**
 * This exception is thrown when, when no valid record is found
 * by the {@link Database_Record} class. It'll report the table's
 * name, its key column and the record's key.
 *
 * @package Database
 * @author  Alessandro Desantis <desa.alessandro@gmail.com>
 */
class DatabaseException extends Exception
{
    /**
     * Contains the table's name.
     *
     * @var    string
     * @access private
     */
    private $_table = '';

    /**
     * Contains the key column's name.
     *
     * @var    string
     * @access private
     */
    private $_column = '';

    /**
     * Contains the record's key (usually an integer).
     *
     * @var    mixed
     * @access private
     */
    private $_value = '';

    /**
     * Formats the exception's message correctly.
     *
     * @param string Table's name
     * @param string Table's key column
     * @param string Record's key
     *
     * @return Database_Exception
     * @access public
     */
    public function __construct($table, $column, $value)
    {
        $this->_table  = $table;
        $this->_column = $column;
        $this->_value  = $value;

        $message = "No record matching the specified criteria ('{$value}' value in '{$column}' column) found in '{$table}' table.";
        parent::__construct($message);
    }

    /**
     * This magic method allows us to provide read-only access
     * to the exception's properties.
     *
     * @param string Property's name
     *
     * @return mixed
     * @access public
     */
    public function __get($property)
    {
        $available = array('table', 'column', 'value');

        if (in_array($property, $available)) {
            if (isset($this->{"_{$property}"})) {
                return $this->{"_{$property}"};
            }
        }
    }
}
?>
